import json
import logging
import pickle
import time

import multiprocessing as mp
import numpy as np

from collections import Counter
from numpy.random import random
from scipy.optimize import (
    minimize,
    basinhopping,
    differential_evolution,
    dual_annealing,
)
from scipy.cluster.hierarchy import linkage, fcluster
from typing import List

import matplotlib.pyplot as plt

plt.switch_backend("agg")

from flare.env import AtomicEnvironment
from flare.gp import GaussianProcess
from flare.gp_algebra import (
    get_like_from_mats,
    get_neg_like_grad,
    get_ky_mat_update,
    _global_training_data,
    _global_training_labels,
    _global_training_structures,
    _global_energy_labels,
    get_Ky_mat,
    get_kernel_vector,
    get_neg_like,
    kernel_distance_mat,
)
from flare.kernels.utils import from_mask_to_args
from flare.output import set_logger
from flare.parameters import Parameters
from flare.struc import Structure
from flare.utils.element_coder import NumpyEncoder, Z_to_element
from typing import Union


class RobustBayesianCommitteeMachine(GaussianProcess):
    """Robust Bayesian Committee Machine.
    Splits up prediction and hyperparameter optimzation over a 'comittee'
    of experts, which gets around the unfavorable nonlinear scaling of
    preidction and training time with training set size. Particularly for
    optimizing hyperparameters (typically the most
    time-consuming individual step of adapting GPs to a new system).

    opt_algorithm GP kwarg supports "dual annealing" and "differential evolution"
    for the RBCMs.

    See:
    Marc Deisenroth, Jun Wei Ng;
    Proceedings of the 32nd International Conference on Machine Learning,
    PMLR 37:1481-1490, 2015, http://proceedings.mlr.press/v37/deisenroth15.html
    for a theory background.

    Args:
        n_experts: number of experts to begin with
        ndata_per_expert: maximum training set size per expert.'
        prior_variance:
        per_expert_parallel: bool to parallelize over experts
        kwargs: same as for GaussianProcess object.
    """

    def __init__(
        self,
        n_experts: int = 1,
        ndata_per_expert: int = 200,
        prior_variance: float = 0.5,
        per_expert_parallel: bool = True,
        **kwargs,
    ):

        self.n_experts = n_experts
        self.prior_variance = prior_variance
        self.log_prior_var = np.log(prior_variance)
        self.ndata_per_expert = ndata_per_expert
        self.per_expert_parallel = per_expert_parallel

        GaussianProcess.__init__(self, **kwargs)
        # Index of which expert is currently addressed
        self.current_expert = 0
        self.reset_container()

    def reset_container(self):
        """
        Perform the instantiation of the RBCM by establishing
        all training data variables as lists: The n-th expert
        will access the n-th index of each of these containers
        to get their own training data and GP parameters.
        :return:
        """

        self.training_data = []
        self.training_labels = []  # Forces acting on central atoms
        self.training_labels_np = []  # np.empty(0, )
        self.n_envs_prev = []  # len(self.training_data)

        # Attributes to accomodate energy labels:
        # self.training_structures = []  # Environments of each structure
        self.energy_labels = []  # Energies of training structures
        self.energy_labels_np = []
        self.all_labels = []

        # Parameters set during training
        self.ky_mat = []
        self.force_block = []
        self.energy_block = []
        self.force_energy_block = []
        self.l_mat = []
        self.alpha = []
        self.ky_mat_inv = []
        self.likelihood = []

        if self.n_experts < 1:
            logger = logging.getLogger(self.logger_name)
            logger.warning("Number of experts is not positive.Setting to 1.")
            self.n_experts = 1
        for i in range(self.n_experts):
            self.add_container()

        self.current_expert = 0

    @staticmethod
    def from_gp(
        gp: GaussianProcess,
        ndata_per_expert: int = 200,
        prior_variance: float = 0.5,
        per_expert_parallel: bool = True,
        post_train: bool = False,
    ) -> "RobustBayesianCommitteeMachine":
        """
        Create an RBCM from a previously existing Gaussian Process with training data.
        Useful for faster hyperparameter optimization!
        """

        rbcm = RobustBayesianCommitteeMachine(
            n_experts=1,
            ndata_per_expert=ndata_per_expert,
            prior_variance=prior_variance,
            per_expert_parallel=per_expert_parallel,
            kernels=gp.kernels,
            hyps=gp.hyps,
            cutoffs=gp.cutoffs,
            hyps_mask=gp.hyps_mask,
            hyp_labels=gp.hyp_labels,
            opt_algorithm=gp.opt_algorithm,
            maxiter=gp.maxiter,
            parallel=gp.parallel,
            per_atom_par=gp.per_atom_par,
            n_sample=gp.n_sample,
            output=gp.output,
            name=gp.name + "_rbcm",
        )

        for env in gp.training_data:
            rbcm.add_one_env(env, force=env.force, train=False)

        if post_train:
            rbcm.train()

        return rbcm

    def check_instantiation(self):
        """
        Runs a series of checks to ensure that the user has not supplied
        contradictory arguments which will result in undefined behavior
        with multiple hyperparameters.
        :return:
        """

        if self.logger_name is None:
            if self.output is None:
                self.logger_name = self.name + "GaussianProcess"
                set_logger(
                    self.logger_name,
                    stream=True,
                    fileout_name=None,
                    verbose=self.verbose,
                )
            else:
                self.logger_name = self.output.basename + "log"
        logger = logging.getLogger(self.logger_name)

        # check whether it's be loaded before
        loaded = False
        if self.name + "_0" in _global_training_labels:
            if (
                _global_training_labels.get(self.name + "_0", None)
                is not self.training_labels_np
            ):
                loaded = True
        if self.name + "_0" in _global_energy_labels:
            if (
                _global_energy_labels.get(self.name + "_0", None)
                is not self.energy_labels_np
            ):
                loaded = True

        if loaded:

            base = f"{self.name}"
            count = 2
            while self.name + "_0" in _global_training_labels and count < 100:
                time.sleep(random())
                self.name = f"{base}_{count}"
                logger.debug(
                    "Specified GP name is present in global memory; "
                    "Attempting to rename the "
                    f"GP instance to {self.name}"
                )
                count += 1
            if self.name + "_0" in _global_training_labels:
                milliseconds = int(round(time.time() * 1000) % 10000000)
                self.name = f"{base}_{milliseconds}"
                logger.debug(
                    "Specified GP name still present in global memory: "
                    f"renaming the gp instance to {self.name}"
                )
            logger.debug(f"Final name of the gp instance is {self.name}")

        self.sync_data()

        self.hyps_mask = Parameters.check_instantiation(
            self.hyps, self.cutoffs, self.kernels, self.hyps_mask
        )

    def find_available_expert(self):
        """
        Determine expert to recieve new data, based on the
        ndata per expert internal value.
        :return:
        """

        expert_id = self.current_expert
        if len(self.training_data[expert_id]) > self.ndata_per_expert:
            self.current_expert += 1
            self.n_experts += 1
            self.add_container()

        return self.current_expert

    def add_container(self):
        """
        Add a new training set for a new expert to access.
        :return:
        """

        self.training_data += [[]]
        self.training_labels += [[]]
        self.training_labels_np += [
            np.empty(
                0,
            )
        ]
        self.n_envs_prev += [0]

        self.training_structures += [[]]  # Environments of each structure
        self.energy_labels += [[]]  # Energies of training structures
        self.energy_labels_np += [
            np.empty(
                0,
            )
        ]
        self.all_labels += [
            np.empty(
                0,
            )
        ]

        self.ky_mat += [None]
        self.force_block += [None]
        self.energy_block += [None]
        self.force_energy_block += [None]
        self.l_mat += [None]
        self.alpha += [None]
        self.ky_mat_inv += [None]

        self.likelihood += [None]

    def update_db(
        self,
        struc: Structure,
        forces: List = None,
        custom_range: List[int] = (),
        energy: float = None,
        expert_id: int = None,
    ):
        """Given a structure and forces, add local environments from the
        structure to the training set of the GP. If energy is given, add the
        entire structure to the training set. expert_id indexes which expert
        gets the new data.

        Args:
            struc (Structure): Input structure. Local environments of atoms
                in this structure will be added to the training set of the GP.
            forces (np.ndarray): Forces on atoms in the structure.
            custom_range (List[int]): Indices of atoms whose local
                environments will be added to the training set of the GP.
            energy (float): Energy of the structure.
        """

        # By default, use all atoms in the structure
        noa = len(struc.positions)
        update_indices = custom_range or list(range(noa))

        # If forces are given, update the environment list.
        for atom in update_indices:
            env_curr = AtomicEnvironment(
                struc, atom, self.cutoffs, cutoffs_mask=self.hyps_mask
            )
            if forces is not None:
                force_curr = np.array(forces[atom])
            else:
                force_curr = env_curr.force

            self.add_one_env(env_curr, force_curr, train=False, expert_id=expert_id)

        # If an energy is given, update the structure list.
        if energy is not None:
            raise NotImplementedError

    def add_one_env(
        self,
        env: AtomicEnvironment,
        force,
        train: bool = False,
        expert_id=None,
        **kwargs,
    ):
        """Add a single local environment to the training set of the GP.
        Args:
            env (AtomicEnvironment): Local environment to be added to the
                training set of the GP.
            force (np.ndarray): Force on the central atom of the local
                environment in the form of a 3-component Numpy array
                containing the x, y, and z components.
            train (bool): If True, the GP is trained after the local
                environment is added.
        """

        if expert_id is None:
            expert_id = self.find_available_expert()

        logger = logging.getLogger(self.logger_name)
        logger.debug(f"add environment to Expert {expert_id}")

        self.training_data[expert_id].append(env)
        self.training_labels[expert_id].append(force)
        self.training_labels_np[expert_id] = np.hstack(self.training_labels[expert_id])

        # update list of all labels
        self.all_labels[expert_id] = np.concatenate(
            (self.training_labels_np[expert_id], self.energy_labels_np[expert_id])
        )

        if train:
            self.train(**kwargs)

    def train(
        self,
        logger_name=None,
        custom_bounds=None,
        grad_tol: float = 1e-4,
        x_tol: float = 1e-5,
        line_steps: int = 20,
        print_progress: Union[bool, str] = False,
        **kwargs,
    ):
        """Train RBCM model on training data. Tunes the
        hyperparameters to maximize the likelihood, then computes L and alpha
        (related to the covariance matrix of the training set).
        Args:
            logger (logging.Logger): logger object specifying where to write the
                progress of the optimization.
            custom_bounds (np.ndarray): Custom bounds on the hyperparameters.
            grad_tol (float): Tolerance of the hyperparameter gradient that
                determines when hyperparameter optimization is terminated.
            x_tol (float): Tolerance on the x values used to decide when
                Nelder-Mead hyperparameter optimization is terminated.
            line_steps (int): Maximum number of line steps for L-BFGS
                hyperparameter optimization.
        """

        verbose = "warning"
        if print_progress:
            verbose = "info"
        if isinstance(print_progress, str) and print_progress.lower() == "debug":
            verbose = "debug"
        if logger_name is None:
            set_logger(
                "gp_algebra",
                stream=False,
                fileout_name="log.gp_algebra",
                verbose=verbose,
            )
            logger_name = "gp_algebra"

        supported_algorithms = [
            "differential evolution",
            "dual annealing",
            "L-BFGS-B",
            "basin hopping",
            "BFGS",
        ]
        if self.opt_algorithm not in supported_algorithms:
            raise ValueError(
                f"Optimization Algorithm {self.opt_algorithm} not "
                f"supported. Please "
                f"choose from {supported_algorithms}"
            )

        disp = False  # print_progress

        if len(self.training_data) == 0 or len(self.training_labels) == 0:
            raise Warning(
                "You are attempting to train a GP with no "
                "training data. Add environments and forces "
                "to the GP and try again."
            )

        self.sync_data()

        x_0 = self.hyps

        opt_algorithm = f"{self.opt_algorithm}"

        args = (
            self.n_experts,
            self.name,
            self.kernel_grad,
            logger_name,
            self.cutoffs,
            self.hyps_mask,
            self.n_cpus,
            self.n_sample,
            self.per_expert_parallel,
        )
        func = rbcm_get_neg_like_grad

        if opt_algorithm in ["differential evolution", "dual annealing"]:
            args0 = (
                self.n_experts,
                self.name,
                self.kernel,
                logger_name,
                self.cutoffs,
                self.hyps_mask,
                self.n_cpus,
                self.n_sample,
                self.per_expert_parallel,
            )
            func0 = rbcm_get_neg_like

        res = None

        if self.bounds is None:
            bounds = np.array([(1e-6, 100)] * len(x_0))
            if self.hyps_mask.get("train_noise", True):
                bounds[-1, 0] = 1e-3
                bounds[-1, 1] = self.prior_variance
        elif custom_bounds is not None:
            bounds = custom_bounds
        else:
            bounds = self.bounds

        if opt_algorithm == "basin hopping":
            minimizer_kwargs = {
                "method": "L-BFGS-B",
                "jac": True,
                "args": args,
                "maxiter": 200,
            }
            res = basinhopping(
                func, x_0, minimizer_kwargs=minimizer_kwargs, niter=self.maxiter
            )

        if opt_algorithm == "differential evolution":
            res = differential_evolution(
                func0, bounds, args=args0, maxiter=self.maxiter, polish=False, **kwargs
            )
            opt_algorithm = "L-BFGS-B"

        if opt_algorithm == "dual annealining":

            res = dual_annealing(
                func0, bounds, args=args0, maxiter=self.maxiter, x0=x_0, **kwargs
            )

        if opt_algorithm == "L-BFGS-B":

            # bound signal noise below to avoid overfitting
            # Catch linear algebra errors and switch to BFGS if necessary
            try:
                res = minimize(
                    func,
                    x_0,
                    args,
                    method="L-BFGS-B",
                    jac=True,
                    bounds=bounds,
                    options={
                        "disp": disp,
                        "gtol": grad_tol,
                        "maxls": line_steps,
                        "maxiter": self.maxiter,
                    },
                )
            except np.linalg.LinAlgError:
                logger = logging.getLogger(self.logger_name)
                logger.warning(
                    "Algorithm for L-BFGS-B failed. Changing to "
                    "BFGS for remainder of run."
                )
                opt_algorithm = "BFGS"

        if opt_algorithm == "BFGS":
            res = minimize(
                func,
                x_0,
                args,
                method="BFGS",
                jac=True,
                options={"disp": disp, "gtol": grad_tol, "maxiter": self.maxiter},
            )

        if res is None:
            raise RuntimeError("Optimization failed for some reason.")
        self.hyps = res.x
        self.set_L_alpha()
        self.total_likelihood = -res.fun
        self.total_likelihood_gradient = -res.jac

        return res

    def check_L_alpha(self):
        """
        For each expert, check that the alpha vector is up to date with the training
        set. If not, update_L_alpha is called.
        """

        self.sync_data()

        # Check that alpha is up to date with training set
        for i in range(self.n_experts):
            size3 = len(self.training_data[i]) * 3

            # If model is empty, then just return
            if size3 == 0:
                return

            if self.alpha[i] is None:
                self.update_L_alpha(i)
            elif size3 > self.alpha[i].shape[0]:
                self.update_L_alpha(i)
            elif size3 != self.alpha[i].shape[0]:
                self.set_L_alpha_part(i)

    def predict(self, x_t: AtomicEnvironment, d: int) -> [float, float]:
        """
        Predict a force component of the central atom of a local environment.

        Performs prediction over each expert and combines results.

        Weights beta_i for experts computed on a per-expert basis following Cao and
        Fleet 2014: https://arxiv.org/abs/1410.7827
        Which Liu et al (https://arxiv.org/pdf/1806.00720.pdf) describe as
        "the difference in differential entropy between the prior and the posterior".

        Args:
            x_t (AtomicEnvironment): Input local environment.
            i (integer):  Force component to predict (1=x, 2=y, 3=z).
        Return:
            (float, float): Mean and epistemic variance of the prediction.
        """

        assert d in [1, 2, 3], "d must be 1, 2 ,or 3."

        # Kernel vector allows for evaluation of atomic environments.
        if self.parallel and not self.per_atom_par:
            n_cpus = self.n_cpus
        else:
            n_cpus = 1

        self.sync_data()

        k_v = []
        for i in range(self.n_experts):
            k_v += [
                get_kernel_vector(
                    f"{self.name}_{i}",
                    self.kernel,
                    self.energy_force_kernel,
                    x_t,
                    d,
                    self.hyps,
                    cutoffs=self.cutoffs,
                    hyps_mask=self.hyps_mask,
                    n_cpus=n_cpus,
                    n_sample=self.n_sample,
                )
            ]

        # Guarantee that alpha is up to date with training set
        self.check_L_alpha()

        # get predictive mean
        variance_rbcm = 0
        mean = 0.0
        var = 0.0
        beta = 0.0  # Represents expert weight

        args = from_mask_to_args(self.hyps, self.cutoffs, self.hyps_mask)

        for i in range(self.n_experts):

            mean_i = np.matmul(self.alpha[i], k_v[i])

            # get predictive variance without cholesky (possibly faster)
            # pass args to kernel based on if mult. hyperparameters in use

            self_kern = self.kernel(x_t, x_t, d, d, *args)
            var_i = self_kern - np.matmul(
                np.matmul(k_v[i].T, self.ky_mat_inv[i]), k_v[i]
            )

            beta_i = 0.5 * (self.log_prior_var - np.log(var_i))  # This expert's weight
            mean += mean_i * beta_i / var_i
            var += beta_i / var_i
            beta += beta_i

        var += (1 - beta) / self.prior_variance
        pred_var = 1.0 / var
        pred_mean = pred_var * mean

        return pred_mean, pred_var

    def predict_force_xyz(self, x_t: AtomicEnvironment) -> ("np.ndarray", "np.ndarray"):
        """
        Simple wrapper to predict all three components of a force in one go.
        :param x_t:
        :return:
        """
        forces = []
        stds = []
        for d in (1, 2, 3):
            force, std = self.predict(x_t, d)
            forces.append(force)
            stds.append(std)
        return np.array(forces), np.array(stds)

    def set_L_alpha(self):
        """
        Set the lower-triangular (L) version of the covariance matrix and alpha vector (
        mapping from  similarity vector of a new training point to output label) for
        each expert.
        :return:
        """

        self.sync_data()

        logger = logging.getLogger(self.logger_name)
        logger.debug("set_L_alpha")

        if self.per_expert_parallel and self.n_cpus > 1:

            time0 = time.time()
            with mp.Pool(processes=self.n_cpus) as pool:
                results = []
                for expert_id in range(self.n_experts):
                    results.append(
                        pool.apply_async(
                            get_Ky_mat,
                            (
                                self.hyps,
                                f"{self.name}_{expert_id}",
                                self.kernel,
                                self.energy_kernel,
                                self.energy_force_kernel,
                                self.energy_noise,
                                self.cutoffs,
                                self.hyps_mask,
                                1,
                                self.n_sample,
                            ),
                        )
                    )
                for i in range(self.n_experts):
                    ky_mat = results[i].get()
                    self.compute_experts_matrices(ky_mat, i)
                pool.close()
                pool.join()
            logger.debug(f"set_L_alpha with per_expert_par {time.time()-time0}")
        else:

            for expert_id in range(self.n_experts):

                logger.debug(f"compute L_alpha for {expert_id}")
                time0 = time.time()

                ky_mat = get_Ky_mat(
                    self.hyps,
                    f"{self.name}_{expert_id}",
                    self.kernel,
                    self.energy_kernel,
                    self.energy_force_kernel,
                    self.energy_noise,
                    self.cutoffs,
                    self.hyps_mask,
                    self.n_cpus,
                    self.n_sample,
                )

                self.compute_experts_matrices(ky_mat, expert_id)
                logger.debug(
                    f"{expert_id} compute_L_alpha {time.time()-time0}"
                    f" {len(self.training_data[expert_id])}"
                )

        for expert_id in range(self.n_experts):
            self.likelihood[expert_id] = get_like_from_mats(
                self.ky_mat[expert_id],
                self.l_mat[expert_id],
                self.alpha[expert_id],
                f"{self.name}_{expert_id}",
            )

        self.total_likelihood = np.sum(self.likelihood)

    def set_L_alpha_part(self, expert_id):
        """
        Invert the covariance matrix, setting L (a lower triangular
        matrix s.t. L L^T = (K + sig_n^2 I)) and alpha, the inverse
        covariance matrix multiplied by the vector of training labels.
        The forces and variances are later obtained using alpha.
        """

        self.sync_experts_data(expert_id)

        if self.per_expert_parallel and self.n_cpus > 1:
            n_cpus = 1
        else:
            n_cpus = self.n_cpus

        ky_mat = get_Ky_mat(
            self.hyps,
            f"{self.name}_{expert_id}",
            self.kernel,
            self.energy_kernel,
            self.energy_force_kernel,
            self.energy_noise,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=n_cpus,
            n_sample=self.n_sample,
        )

        self.compute_experts_matrices(ky_mat, expert_id)

        self.likelihood[expert_id] = get_like_from_mats(
            self.ky_mat[expert_id],
            self.l_mat[expert_id],
            self.alpha[expert_id],
            f"{self.name}_{expert_id}",
        )

    def sync_data(self):
        """
        Ensure data is correctly indexed in global memory
        :return:
        """
        for i in range(self.n_experts):
            self.sync_experts_data(i)

    def unsync_experts_data(self, expert_id: int):
        """ Reset global variables. """
        if len(self.training_data) > expert_id:
            _global_training_data.pop(f"{self.name}_{expert_id}", None)
            _global_training_labels.pop(f"{self.name}_{expert_id}", None)
            # _global_training_structures.pop(f"{self.name}_{expert_id}",None)
            _global_energy_labels.pop(f"{self.name}_{expert_id}", None)

    def sync_experts_data(self, expert_id: int):
        """ Reset global variables. """
        if len(self.training_data) > expert_id:
            _global_training_data[f"{self.name}_{expert_id}"] = self.training_data[
                expert_id
            ]
            _global_training_labels[
                f"{self.name}_{expert_id}"
            ] = self.training_labels_np[expert_id]
            _global_training_structures[
                f"{self.name}_{expert_id}"
            ] = self.training_structures[expert_id]
            _global_energy_labels[f"{self.name}_{expert_id}"] = self.energy_labels_np[
                expert_id
            ]

    def update_L_alpha(self, expert_id: int):
        """
        Update the GP's L matrix and alpha vector without recalculating
        the entire covariance matrix K.
        """

        # Set L matrix and alpha if set_L_alpha has not been called yet
        if self.l_mat[expert_id] is None or np.array(
            self.ky_mat[expert_id]
        ) is np.array(None):
            self.set_L_alpha_part(expert_id)
            return

        self.sync_experts_data(expert_id)

        ky_mat = get_ky_mat_update(
            self.ky_mat[expert_id],
            self.n_envs_prev[expert_id],
            self.hyps,
            f"{self.name}_{expert_id}",
            self.kernel,
            self.energy_kernel,
            self.energy_force_kernel,
            self.energy_noise,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=self.n_cpus,
            n_sample=self.n_sample,
        )

        self.compute_experts_matrices(ky_mat, expert_id)

    def compute_experts_matrices(self, ky_mat, expert_id):
        """
        When covariance matrix is known, reconstruct other matrices.
        Used in re-loading large GPs.
        :return:
        """

        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.all_labels[expert_id])

        self.ky_mat[expert_id] = ky_mat
        self.l_mat[expert_id] = l_mat
        self.alpha[expert_id] = alpha
        self.ky_mat_inv[expert_id] = ky_mat_inv
        self.n_envs_prev[expert_id] = len(self.training_data[expert_id])

    def redistribute_training_data(self, figure_name="default"):
        """ redistribute data """

        joint_data = []
        joint_structures = []
        joint_labels = []
        for i in range(self.n_experts):
            joint_data += self.training_data[i]
            joint_labels += self.training_labels[i]
            self.unsync_experts_data(i)

        self.n_experts = 1
        self.reset_container()

        _global_training_data[f"{self.name}_join"] = joint_data
        _global_training_structures[f"{self.name}_join"] = joint_structures

        kmat = kernel_distance_mat(
            self.hyps,
            self.name + "_join",
            self.energy_kernel,
            self.cutoffs,
            self.hyps_mask,
            self.n_cpus,
            self.n_sample,
        )
        for i in range(kmat.shape[0]):
            norm = np.sqrt(kmat[i, i])
            kmat[i, :] /= norm
            kmat[:, i] /= norm
        kmat = np.max(kmat) - kmat
        print(np.min(kmat), np.max(kmat))

        plot_mat(kmat, f"{figure_name}_unsorted_dmat.png")
        # np.save(f"{figure_name}_unsorted_dmat.npy", kmat)

        iu1 = np.triu_indices(kmat.shape[0], 1)
        upper_triang = kmat[iu1]
        Z = linkage(upper_triang, method="average")
        K = fcluster(Z, t=0.5, criterion="distance")

        ncluster = int(np.max(K))
        new_indices = []
        sizes = []
        for i in range(ncluster):
            index = np.where(K == (i + 1))[0]
            new_indices += [index]
            sizes += [len(index)]
        new_indices = np.hstack(new_indices)

        plot_mat(sort_matrix(kmat, new_indices), f"{figure_name}_sorted_dmat.png")
        # np.save(f"{figure_name}_sorted_dmat.npy", kmat)

        for i in new_indices:
            self.add_one_env(
                joint_data[i], joint_labels[i], train=False, expert_id=None
            )

        del _global_training_data[f"{self.name}_join"]
        del _global_training_structures[f"{self.name}_join"]
        del joint_data
        del joint_structures
        del joint_labels

        self.set_L_alpha()

        return kmat

    def compute_join_dist_mat(self, list_expert_id):
        """ Reset global variables. """
        joint_data = []
        joint_structures = []
        for i in list_expert_id:
            joint_data += self.training_data[i]
            # joint_structures += self.training_structures[expert_id]
        _global_training_data[f"{self.name}_join"] = joint_data
        _global_training_structures[f"{self.name}_join"] = joint_labels
        kmat = kernel_distance_mat(
            self.hyps,
            self.name + "_join",
            self.energy_kernel,
            self.cutoffs,
            self.hyps_mask,
            self.n_cpus,
            self.n_sample,
        )
        for i in range(kmat.shape[0]):
            norm = np.sqrt(kmat[i, i])
            kmat[i, :] /= norm
            kmat[:, i] /= norm
        print(np.min(kmat), np.max(kmat))
        kmat = np.max(kmat) - kmat

        del _global_training_data[f"{self.name}_join"]
        del _global_training_structures[f"{self.name}_join"]
        del joint_data
        del joint_structures
        return kmat

    @property
    def training_statistics(self) -> dict:
        """
        Return dict with statistics about the current training data by expert.
        Useful for quickly summarizing info about the RBCM.
        :return:
        """

        data = {}

        # Count all of the present species in the atomic env. data
        present_species = []
        data["N"] = 0
        for i in range(self.n_experts):
            data["N"] += self.n_envs_prev[i]
            data[f"N_{i}"] = self.n_envs_prev[i]
            for env, _ in zip(self.training_data[i], self.training_labels[i]):
                present_species.append(
                    Z_to_element(env.structure.coded_species[env.atom])
                )

        # Summarize the relevant information
        data["species"] = list(set(present_species))
        data["envs_by_species"] = dict(Counter(present_species))

        return data

    def write_model(self, name: str, format: str = "pickle"):
        """
        Write model in a variety of formats to a file for later re-use.
        Currently only pickle / binary is supported for RBCMs.
        Args:
            name (str): Output name.
            format (str): Output format.
        """

        supported_formats = ["pickle", "binary"]

        for detect in ["json", "pickle", "binary"]:
            if detect in name.lower():
                format = detect
                break

        format = format.lower()

        if format == "pickle" or format == "binary":
            if ".pickle" != name[-7:]:
                name += ".pickle"
            with open(f"{name}", "wb") as f:
                pickle.dump(self, f)

        else:
            raise ValueError(
                f"Output format {format} not supported: try "
                f"from {supported_formats}"
            )

    def get_full_gp(self) -> GaussianProcess:
        """
        "Flatten" an RBCM into a standard Gaussian Process.

        "I think the people of this country have had enough of experts."
                                                    - Michael Gove
        :return:
        """

        gp_model = GaussianProcess(**self.__dict__)
        gp_model.training_data = []
        gp_model.training_labels = []
        gp_model.name = "rbcm_derived_gp"
        for i in range(self.n_experts):
            gp_model.training_data += self.training_data[i]
            gp_model.training_labels += self.training_labels[i]
            gp_model.training_labels_np = np.hstack(gp_model.training_labels)
            gp_model.all_labels = np.hstack(
                (gp_model.training_labels_np, gp_model.energy_labels_np)
            )
        gp_model.sync_data()
        return gp_model

    @staticmethod
    def from_file(filename: str, format: str = ""):
        """
        One-line convenience method to load a GP from a file stored using
        write_file
        Args:
            filename (str): path to GP model
            format (str): json or pickle if format is not in filename
        :return:
        """

        supported_formats = ["pickle", "binary"]

        if ".json" in filename or "json" in format:
            raise ValueError(
                "Output format not supported: try from {}".format(supported_formats)
            )

        if "binary" in format.lower() or "binary" in filename.lower():
            format = "pickle"

        elif ".pickle" in filename or "pickle" in format:
            with open(filename, "rb") as f:

                gp_model = pickle.load(f)

                GaussianProcess.backward_arguments(gp_model.__dict__, gp_model.__dict__)

                GaussianProcess.backward_attributes(gp_model.__dict__)

        else:
            raise ValueError(
                "Warning: Format unspecieified or file is not .json or .pickle format."
            )

        # TODO, be careful of this one
        gp_model.sync_data()

        return gp_model

    def __str__(self):
        """String representation of the GP model."""

        thestr = "GaussianProcess Object\n"
        thestr += f"Number of cpu cores: {self.n_cpus}\n"
        thestr += f"Kernel: {self.kernels}\n"
        thestr += f"Training points: {sum([len(x) for x in self.training_data])}\n"
        thestr += f"Cutoffs: {self.cutoffs}\n"
        thestr += f"Number of Experts: {len(self.training_data)}\n"


        thestr += f"Number of hyperparameters: {len(self.hyps)}\n"
        thestr += f"Hyperparameters_array: {str(self.hyps)}\n"
        thestr += "Hyperparameters: \n"
        if self.hyp_labels is None:
            # Put unlabeled hyperparameters on one line
            thestr = thestr[:-1]
            thestr += str(self.hyps) + "\n"
        else:
            for hyp, label in zip(self.hyps, self.hyp_labels):
                thestr += f"{label}: {hyp} \n"

        for k in self.hyps_mask:
            thestr += f"Hyps_mask {k}: {self.hyps_mask[k]} \n"

        return thestr


def rbcm_get_neg_like_grad(
    hyps,
    n_experts,
    name,
    kernel_grad,
    logger_name,
    cutoffs,
    hyps_mask,
    n_cpus,
    n_sample,
    per_expert_parallel,
):

    neg_like = 0
    neg_like_grad = None

    logger = logging.getLogger(logger_name)
    time0 = time.time()
    if per_expert_parallel and n_cpus > 1:

        with mp.Pool(processes=n_cpus) as pool:

            results = []
            for i in range(n_experts):
                results.append(
                    pool.apply_async(
                        get_neg_like_grad,
                        (
                            hyps,
                            f"{name}_{i}",
                            kernel_grad,
                            logger_name,
                            cutoffs,
                            hyps_mask,
                            1,
                            n_sample,
                        ),
                    )
                )
            for i in range(n_experts):
                chunk = results[i].get()
                neg_like_, neg_like_grad_ = chunk
                neg_like += neg_like_
                if neg_like_grad is None:
                    neg_like_grad = neg_like_grad_
                else:
                    neg_like_grad += neg_like_grad_
            pool.close()
            pool.join()
    else:
        for i in range(n_experts):
            neg_like_, neg_like_grad_ = get_neg_like_grad(
                hyps,
                f"{name}_{i}",
                kernel_grad,
                logger_name,
                cutoffs,
                hyps_mask,
                n_cpus,
                n_sample,
            )
            neg_like += neg_like_
            if neg_like_grad is None:
                neg_like_grad = neg_like_grad_
            else:
                neg_like_grad += neg_like_grad_

    logger.info("")
    logger.info(f"Hyperparameters: {list(hyps)}")
    logger.info(f"Total Likelihood: {-neg_like}")
    logger.info(f"Total Likelihood Gradient: {list(neg_like_grad)}")
    logger.info(f"One step {time.time()-time0}")

    ohyps, label = Parameters.get_hyps(hyps_mask, hyps, constraint=False, label=True)
    if label:
        logger.info(f"oHyp_array: {list(ohyps)}")
        for i, l in enumerate(label):
            logger.info(f"oHyp {l}: {ohyps[i]}")

    return neg_like, neg_like_grad


def rbcm_get_neg_like(
    hyps,
    n_experts,
    name,
    force_kernel,
    logger_name,
    cutoffs,
    hyps_mask,
    n_cpus,
    n_sample,
    per_expert_parallel,
):

    neg_like = 0
    neg_like_grad = None

    logger = logging.getLogger(logger_name)
    time0 = time.time()
    if per_expert_parallel and n_cpus > 1:

        with mp.Pool(processes=n_cpus) as pool:

            results = []
            for i in range(n_experts):
                results.append(
                    pool.apply_async(
                        get_neg_like,
                        (
                            hyps,
                            f"{name}_{i}",
                            force_kernel,
                            logger_name,
                            cutoffs,
                            hyps_mask,
                            1,
                            n_sample,
                        ),
                    )
                )
            for i in range(n_experts):
                chunk = results[i].get()
                neg_like_ = chunk
                neg_like += neg_like_
            pool.close()
            pool.join()
    else:
        for i in range(n_experts):
            neg_like_ = get_neg_like(
                hyps,
                f"{name}_{i}",
                force_kernel,
                logger_name,
                cutoffs,
                hyps_mask,
                n_cpus,
                n_sample,
            )
            neg_like += neg_like_

    logger.info("")
    logger.info(f"Hyperparameters: {list(hyps)}")
    logger.info(f"Total Likelihood: {-neg_like}")
    logger.info(f"One step {time.time()-time0}")

    ohyps, label = Parameters.get_hyps(hyps_mask, hyps, constraint=False, label=True)
    if label:
        logger.info(f"oHyp_array: {list(ohyps)}")
        for i, l in enumerate(label):
            logger.info(f"oHyp {l}: {ohyps[i]}")

    return neg_like


def sort_matrix(matrix, new_order):
    n_obs = matrix.shape[0]
    new_mat = np.zeros_like(matrix)
    for i in range(n_obs):
        for j in range(n_obs):
            new_mat[i, j] = matrix[new_order[i], new_order[j]]
    return new_mat


def plot_mat(distance_mat, filename):
    fig = plt.figure()
    plt.imshow(distance_mat)
    plt.title("distance matrix (1-normalized(<en|en>)")
    plt.tight_layout()
    fig.savefig(filename)
    plt.close()



"""
This module is used to call Quantum Espresso simulation and parse its output
The user need to supply a complete input script with single-point scf
calculation, CELL_PARAMETERS, ATOMIC_POSITIONS, nat, ATOMIC_SPECIES
arguments.  It is case sensitive. and the nat line should be the first
argument of the line it appears. The user can also opt to the ASE interface instead.

This module will copy the input template to a new file with "_run" suffix,
edit the atomic coordination in the ATOMIC_POSITIONS block and run the similation with the parallel set up given.
"""

import os
from subprocess import call
import time
import numpy as np
from flare import struc
from typing import List

name = "QE"


def run_dft_par(
    dft_input,
    structure,
    dft_loc,
    n_cpus=1,
    dft_out="pwscf.out",
    npool=None,
    mpi="mpi",
    **dft_kwargs,
):
    """run DFT calculation with given input template
    and atomic configurations. if n_cpus == 1, it executes serial run.

    :param dft_input: input template file name
    :param structure: atomic configuration
    :param dft_loc:   relative/absolute executable of the DFT code
    :param n_cpus:   # of CPU for mpi
    :param dft_out:   output file name
    :param npool:     not used
    :param mpi:       not used
    :param **dft_wargs: not used
    :return: forces
    """

    newfilename = edit_dft_input_positions(dft_input, structure)

    if npool is None:
        dft_command = f"{dft_loc} -i {newfilename}"
    else:
        dft_command = f"{dft_loc} -nk {npool} -i {newfilename}"

    if n_cpus > 1:
        if mpi == "mpi":
            dft_command = f"mpirun -np {n_cpus} {dft_command}"
        else:
            dft_command = f"srun -n {n_cpus} --mpi=pmi2 {dft_command}"

    with open(dft_out, "w+") as fout:
        call(dft_command.split(), stdout=fout)

    os.remove(newfilename)

    return parse_dft_forces(dft_out)


def run_dft_en_par(dft_input, structure, dft_loc, n_cpus):
    """run DFT calculation with given input template
    and atomic configurations. This function is not used atm

    if n_cpus == 1, it executes serial run.

    :param dft_input: input template file name
    :param structure: atomic configuration
    :param dft_loc:   relative/absolute executable of the DFT code
    :param n_cpus:   # of CPU for mpi
    :param dft_out:   output file name
    :param npool:     not used
    :param mpi:       not used
    :param **dft_wargs: not used
    :return: forces, energy
    """

    run_qe_path = dft_input
    edit_dft_input_positions(run_qe_path, structure)
    qe_command = "mpirun -np {n_cpus} {dft_loc} -i {run_qe_path}"

    with open("pwscf.out", "w+") as fout:
        call(qe_command.split(), stdout=fout)

    forces, energy = parse_dft_forces_and_energy("pwscf.out")

    return forces, energy


def run_dft_en_npool(qe_input, structure, dft_loc, npool):
    run_qe_path = qe_input
    edit_dft_input_positions(run_qe_path, structure)
    qe_command = "mpirun {0} -npool {1} < {2} > {3}".format(
        dft_loc, npool, run_qe_path, "pwscf.out"
    )
    call(qe_command, shell=True)
    forces, energy = parse_dft_forces_and_energy("pwscf.out")

    return forces, energy


def parse_dft_input(dft_input: str):
    """parse the input to get information of atomic configuration

    :param dft_input: input file name
    :return: positions, species, cell, masses
    """

    positions = []
    species = []
    cell = []

    with open(dft_input) as f:
        lines = f.readlines()

    # Find the cell and positions in the output file
    cell_index = None
    positions_index = None
    nat = None
    species_index = None

    for i, line in enumerate(lines):
        if "CELL_PARAMETERS" in line:
            cell_index = int(i + 1)
        if "ATOMIC_POSITIONS" in line:
            positions_index = int(i + 1)
        if "nat" in line:
            nat = int(line.split("=")[1])
        if "ATOMIC_SPECIES" in line:
            species_index = int(i + 1)

    assert cell_index is not None, "Failed to find cell in input"
    assert positions_index is not None, "Failed to find positions in input"
    assert nat is not None, "Failed to find number of atoms in input"
    assert species_index is not None, "Failed to find atomic species in input"

    # Load cell
    for i in range(cell_index, cell_index + 3):
        cell_line = lines[i].strip()
        cell.append(np.fromstring(cell_line, sep=" "))
    cell = np.array(cell)

    # Check cell IO
    assert len(cell) != 0, "Cell failed to load"
    assert np.shape(cell) == (3, 3), "Cell failed to load correctly"

    # Load positions
    for i in range(positions_index, positions_index + nat):
        line_string = lines[i].strip().split()
        species.append(line_string[0])

        pos_string = " ".join(line_string[1:4])

        positions.append(np.fromstring(pos_string, sep=" "))
    # Check position IO
    assert positions != [], "Positions failed to load"
    positions = np.array(positions)

    # see conversions.nb for conversion from amu to md units
    massconvert = 0.000103642695727
    masses = {}
    for i in range(species_index, species_index + len(set(species))):
        # Expects lines of format like: H 1.0 H_pseudo_name.ext
        line = lines[i].strip().split()
        masses[line[0]] = float(line[1]) * massconvert

    return positions, species, cell, masses


def dft_input_to_structure(dft_input: str):
    """Parses a qe input and returns the atoms in the
    file as a Structure object

    :param dft_input: QE Input file to parse
    :return: class Structure
    """
    positions, species, cell, masses = parse_dft_input(dft_input)
    _, coded_species = struc.get_unique_species(species)
    return struc.Structure(
        positions=positions,
        species=coded_species,
        cell=cell,
        mass_dict=masses,
        species_labels=species,
    )


def edit_dft_input_positions(dft_input: str, structure):
    """
    Write the current configuration of the OTF structure to the
    qe input file

    :param dft_input: dft input file name
    :param structure: atomic structure to compute
    :return: the name of the edited file
    """

    with open(dft_input, "r") as f:
        lines = f.readlines()

    file_pos_index = None
    cell_index = None
    nat = None
    for i, line in enumerate(lines):
        if "ATOMIC_POSITIONS" in line:
            file_pos_index = int(i + 1)
        if "CELL_PARAMETERS" in line:
            cell_index = int(i + 1)
        # Load nat into variable then overwrite it with new nat
        if "nat" in line:
            nat = int(line.split("=")[1])
            nat_index = int(i)
            lines[nat_index] = "nat = " + str(structure.nat) + "\n"

    assert file_pos_index is not None, "Failed to find positions in input"
    assert cell_index is not None, "Failed to find cell in input"
    assert nat is not None, "Failed to find nat in input"

    # TODO Catch case where the punchout structure has more atoms than the
    # original structure

    for pos_index, line_index in enumerate(
        range(file_pos_index, file_pos_index + structure.nat)
    ):
        pos_string = " ".join(
            [
                structure.species_labels[pos_index],
                str(structure.positions[pos_index][0]),
                str(structure.positions[pos_index][1]),
                str(structure.positions[pos_index][2]),
            ]
        )
        if line_index < len(lines):
            lines[line_index] = str(pos_string + "\n")
        else:
            lines.append(str(pos_string + "\n"))

    # TODO current assumption: if there is a new structure, then the new
    # structure has fewer atoms than the  previous one. If we are always
    # 'editing' a version of the larger structure than this will be okay with
    # the punchout method.
    for line_index in range(file_pos_index + structure.nat, file_pos_index + nat):
        lines[line_index] = ""

    lines[cell_index] = " ".join([str(x) for x in structure.vec1]) + "\n"
    lines[cell_index + 1] = " ".join([str(x) for x in structure.vec2]) + "\n"
    lines[cell_index + 2] = " ".join([str(x) for x in structure.vec3]) + "\n"

    newfilename = dft_input + "_run"

    with open(newfilename, "w") as f:
        for line in lines:
            f.write(line)

    return newfilename


def parse_dft_forces(outfile: str):
    """
    Get forces from a pwscf file in eV/A

    :param outfile: str, Path to pwscf output file
    :return: list[nparray] , List of forces acting on atoms
    """
    forces = []
    total_energy = np.nan

    with open(outfile, "r") as outf:
        for line in outf:
            if line.lower().startswith("!    total energy"):
                total_energy = float(line.split()[-2])

            if line.find("force") != -1 and line.find("atom") != -1:
                line = line.split("force =")[-1]
                line = line.strip()
                line = line.split(" ")
                line = [x for x in line if x != ""]
                temp_forces = []
                for x in line:
                    temp_forces.append(float(x))
                forces.append(np.array(list(temp_forces)))

    assert (
        total_energy != np.nan
    ), "Quantum ESPRESSO parser failed to read the file {}. Run failed.".format(outfile)

    # Convert from ry/au to ev/angstrom
    conversion_factor = 25.71104309541616

    forces = [conversion_factor * force for force in forces]
    forces = np.array(forces)

    return forces


def parse_dft_forces_and_energy(outfile: str):
    """
    Get forces from a pwscf file in eV/A

    :param outfile: str, Path to pwscf output file
    :return: list[nparray] , List of forces acting on atoms
    """
    forces = []
    total_energy = np.nan

    with open(outfile, "r") as outf:
        for line in outf:
            if line.lower().startswith("!    total energy"):
                total_energy = float(line.split()[-2])

            if line.find("force") != -1 and line.find("atom") != -1:
                line = line.split("force =")[-1]
                line = line.strip()
                line = line.split(" ")
                line = [x for x in line if x != ""]
                temp_forces = []
                for x in line:
                    temp_forces.append(float(x))
                forces.append(np.array(list(temp_forces)))

    assert (
        total_energy != np.nan
    ), "Quantum ESPRESSO parser failed to read the file {}. Run failed.".format(outfile)

    # Convert from ry/au to ev/angstrom
    conversion_factor = 25.71104309541616

    forces = [conversion_factor * force for force in forces]
    forces = np.array(forces)

    return forces, total_energy



from ase.md.npt import NPT


class NPT_mod(NPT):
    def stresscalculator(self):
        stress = self.atoms.get_stress(include_ideal_gas=True)

        # Clear the calculator before forces are computed on new positions.
        self.atoms.calc.results = {}

        return stress



import numpy as np
from numba import njit
from math import floor, ceil

from typing import List

from flare.struc import Structure
from flare.utils.element_coder import Z_to_element

from flare.mgp.mapxb import MapXbody, SingleMapXbody
from flare.mgp.grid_kernels import grid_kernel, self_kernel

from flare.kernels.utils import from_mask_to_args


class Map3body(MapXbody):
    def __init__(self, **kwargs):

        self.kernel_name = "threebody"
        self.singlexbody = SingleMap3body
        self.bodies = 3
        self.pred_perm = [[0, 1, 2], [1, 0, 2]]
        self.spc_perm = [[0, 1, 2], [0, 2, 1]]
        self.num_lmp_maps = 0
        super().__init__(**kwargs)

    def build_bond_struc(self, species_list):
        """
        build a bond structure, used in grid generating
        """

        # 2 body (2 atoms (1 bond) config)
        self.spc = []
        N_spc = len(species_list)
        self.num_lmp_maps = N_spc ** 3
        for spc1 in species_list:
            for spc2 in species_list:
                for spc3 in species_list:
                    species = [spc1, spc2, spc3]
                    self.spc.append(species)

    def get_arrays(self, atom_env):

        spcs, comp_r, comp_xyz = get_triplets(
            atom_env.ctype,
            atom_env.etypes,
            atom_env.bond_array_3,
            atom_env.cross_bond_inds,
            atom_env.cross_bond_dists,
            atom_env.triplet_counts,
        )

        return spcs, comp_r, comp_xyz

    def find_map_index(self, spc):
        return self.spc.index(spc)


class SingleMap3body(SingleMapXbody):
    def __init__(self, **kwargs):
        """
        Build 3-body MGP

        """

        self.bodies = 3
        self.grid_dim = 3
        self.kernel_name = "threebody"
        self.pred_perm = [[0, 1, 2], [1, 0, 2]]

        super().__init__(**kwargs)

        # initialize bounds
        self.set_bounds(None, None)

        spc = self.species
        self.species_code = "_".join([Z_to_element(spc) for spc in self.species])
        self.kv3name = f"kv3_{self.species_code}"

    def set_bounds(self, lower_bound, upper_bound):
        if self.auto_lower:
            if isinstance(lower_bound, float):
                self.bounds[0] = np.ones(3) * lower_bound
            else:
                self.bounds[0] = lower_bound
        if self.auto_upper:
            if isinstance(upper_bound, float):
                self.bounds[1] = np.ones(3) * upper_bound
            else:
                self.bounds[1] = upper_bound

    def construct_grids(self):
        """
        Return:
            An array of shape (n_grid, 3)
        """
        # build grids in each dimension
        triplets = []
        for d in range(3):
            bonds = np.linspace(
                self.bounds[0][d], self.bounds[1][d], self.grid_num[d], dtype=np.float64
            )
            triplets.append(bonds)

        # concatenate into one array: n_grid x 3
        mesh = np.meshgrid(*triplets, indexing="ij")
        del triplets

        mesh_list = []
        n_grid = np.prod(self.grid_num)
        for d in range(3):
            mesh_list.append(np.reshape(mesh[d], n_grid))

        mesh_list = np.array(mesh_list).T

        return mesh_list

    def grid_cutoff(self, triplets, r_cut, coords, derivative, cutoff_func):
        return bonds_cutoff(triplets, r_cut, coords, derivative, cutoff_func)

    def get_grid_kernel(self, kern_type, data, kernel_info, *grid_arrays):
        c2 = self.species[0]
        etypes2 = np.array(self.species[1:])

        _, cutoffs, hyps, hyps_mask = kernel_info
        hyps, r_cut = get_hyps_for_kern(hyps, cutoffs, hyps_mask, c2, etypes2)
        return grid_kernel(
            data,
            self.bodies,
            kern_type,
            get_bonds_for_kern,
            bonds_cutoff,
            c2,
            etypes2,
            hyps,
            r_cut,
            *grid_arrays,
        )

    def get_self_kernel(self, kernel_info, *grid_arrays):
        c2 = self.species[0]
        etypes2 = np.array(self.species[1:])

        _, cutoffs, hyps, hyps_mask = kernel_info
        hyps, r_cut = get_hyps_for_kern(hyps, cutoffs, hyps_mask, c2, etypes2)
        return self_kernel(
            self.bodies, get_permutations, c2, etypes2, hyps, r_cut, *grid_arrays
        )


# -----------------------------------------------------------------------------
#                               Functions
# -----------------------------------------------------------------------------


def bonds_cutoff(triplets, r_cut, coords, derivative, cutoff_func):
    dfj_list = np.zeros((len(triplets), 3), dtype=np.float64)

    if derivative:
        for d in range(3):
            inds = np.arange(3) * 3 + d
            f0, df0 = cutoff_func(r_cut, triplets, coords[:, inds])
            dfj = (
                df0[:, 0] * f0[:, 1] * f0[:, 2]
                + f0[:, 0] * df0[:, 1] * f0[:, 2]
                + f0[:, 0] * f0[:, 1] * df0[:, 2]
            )
            dfj_list[:, d] = dfj
    else:
        f0, _ = cutoff_func(r_cut, triplets, 0)  # (n_grid, 3)

    fj = f0[:, 0] * f0[:, 1] * f0[:, 2]  # (n_grid,)
    fj = np.expand_dims(fj, axis=1)

    return fj, dfj_list


# TODO: move this func to Parameters class
def get_hyps_for_kern(hyps, cutoffs, hyps_mask, c2, etypes2):
    """
    Args:
        data: a single env of a list of envs
    """

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    if len(args) == 2:
        hyps, cutoffs = args
        r_cut = cutoffs[1]

    else:
        (
            cutoff_2b,
            cutoff_3b,
            cutoff_mb,
            nspec,
            spec_mask,
            nbond,
            bond_mask,
            ntriplet,
            triplet_mask,
            ncut3b,
            cut3b_mask,
            nmb,
            mb_mask,
            sig2,
            ls2,
            sig3,
            ls3,
            sigm,
            lsm,
        ) = args

        bc1 = spec_mask[c2]
        bc2 = spec_mask[etypes2[0]]
        bc3 = spec_mask[etypes2[1]]
        ttype = triplet_mask[nspec * nspec * bc1 + nspec * bc2 + bc3]
        ls = ls3[ttype]
        sig = sig3[ttype]
        r_cut = cutoff_3b
        hyps = [sig, ls]

    return hyps, r_cut


@njit
def get_triplets(
    ctype, etypes, bond_array, cross_bond_inds, cross_bond_dists, triplets
):
    exist_species = []
    tris = []
    tri_dir = []

    for m in range(bond_array.shape[0]):
        r1 = bond_array[m, 0]
        c1 = bond_array[m, 1:]
        spc1 = etypes[m]

        for n in range(triplets[m]):
            ind1 = cross_bond_inds[m, m + n + 1]
            r2 = bond_array[ind1, 0]
            c2 = bond_array[ind1, 1:]
            spc2 = etypes[ind1]

            c12 = np.sum(c1 * c2)
            r12 = np.sqrt(r1 ** 2 + r2 ** 2 - 2 * r1 * r2 * c12)

            if spc1 <= spc2:
                spcs = [ctype, spc1, spc2]
                triplet = np.array([r1, r2, r12])
                coord = [c1, c2, np.zeros(3)]
            else:
                spcs = [ctype, spc2, spc1]
                triplet = np.array([r2, r1, r12])
                coord = [c2, c1, np.zeros(3)]

            if spcs not in exist_species:
                exist_species.append(spcs)
                tris.append([triplet])
                tri_dir.append([coord])
            else:
                k = exist_species.index(spcs)
                tris[k].append(triplet)
                tri_dir[k].append(coord)

    return exist_species, tris, tri_dir


@njit
def get_permutations(c1, etypes1, c2, etypes2):
    ei1 = etypes1[0]
    ei2 = etypes1[1]
    ej1 = etypes2[0]
    ej2 = etypes2[1]

    perms = []
    if c1 == c2:
        if (ei1 == ej1) and (ei2 == ej2):
            perms.append([0, 1, 2])
        if (ei1 == ej2) and (ei2 == ej1):
            perms.append([1, 0, 2])
    if c1 == ej1:
        if (ei1 == ej2) and (ei2 == c2):
            perms.append([1, 2, 0])
        if (ei1 == c2) and (ei2 == ej2):
            perms.append([0, 2, 1])
    if c1 == ej2:
        if (ei1 == ej1) and (ei2 == c2):
            perms.append([2, 1, 0])
        if (ei1 == c2) and (ei2 == ej1):
            perms.append([2, 0, 1])
    return perms


def get_bonds_for_kern(env, c2, etypes2):
    return get_triplets_for_kern_jit(
        env.bond_array_3,
        env.ctype,
        env.etypes,
        env.cross_bond_inds,
        env.cross_bond_dists,
        env.triplet_counts,
        c2,
        etypes2,
    )


@njit
def get_triplets_for_kern_jit(
    bond_array_1,
    c1,
    etypes1,
    cross_bond_inds_1,
    cross_bond_dists_1,
    triplets_1,
    c2,
    etypes2,
):

    # triplet_list = np.empty((0, 6), dtype=np.float64)
    triplet_list = []

    ej1 = etypes2[0]
    ej2 = etypes2[1]

    all_spec = [c2, ej1, ej2]
    if c1 in all_spec:
        c1_ind = all_spec.index(c1)
        ind_list = [0, 1, 2]
        ind_list.remove(c1_ind)
        all_spec.remove(c1)

        for m in range(bond_array_1.shape[0]):
            two_inds = [ind_list[0], ind_list[1]]

            ri1 = bond_array_1[m, 0]
            ci1 = bond_array_1[m, 1:]
            ei1 = etypes1[m]

            two_spec = [all_spec[0], all_spec[1]]
            if ei1 in two_spec:

                ei1_ind = ind_list[0] if ei1 == two_spec[0] else ind_list[1]
                two_spec.remove(ei1)
                two_inds.remove(ei1_ind)
                one_spec = two_spec[0]
                ei2_ind = two_inds[0]

                for n in range(triplets_1[m]):
                    ind1 = cross_bond_inds_1[m, m + n + 1]
                    ei2 = etypes1[ind1]
                    if ei2 == one_spec:

                        ri2 = bond_array_1[ind1, 0]
                        ci2 = bond_array_1[ind1, 1:]

                        ri3 = cross_bond_dists_1[m, m + n + 1]
                        ci3 = np.zeros(3)

                        perms = get_permutations(c1, np.array([ei1, ei2]), c2, etypes2)

                        tri = np.array([ri1, ri2, ri3])
                        crd1 = np.array([ci1[0], ci2[0], ci3[0]])
                        crd2 = np.array([ci1[1], ci2[1], ci3[1]])
                        crd3 = np.array([ci1[2], ci2[2], ci3[2]])

                        # append permutations
                        nperm = len(perms)
                        for iperm in range(nperm):
                            perm = perms[iperm]
                            tricrd = np.take(tri, perm)
                            crd1_p = np.take(crd1, perm)
                            crd2_p = np.take(crd2, perm)
                            crd3_p = np.take(crd3, perm)
                            crd_p = np.vstack((crd1_p, crd2_p, crd3_p))
                            tricrd = np.hstack(
                                (tricrd, crd_p[:, 0], crd_p[:, 1], crd_p[:, 2])
                            )
                            triplet_list.append(tricrd)
                            # tricrd = np.expand_dims(tricrd, axis=0)
                            # triplet_list = np.vstack((triplet_list, tricrd))

    return triplet_list



import numpy as np
from flare.kernels.kernels import force_helper, force_energy_helper, grad_helper
from numba import njit
from flare.env import AtomicEnvironment
from typing import Callable
import flare.kernels.cutoffs as cf
from math import exp


class TwoBodyKernel:
    def __init__(
        self,
        hyperparameters: "ndarray",
        cutoff: float,
        cutoff_func: Callable = cf.quadratic_cutoff,
    ):
        self.hyperparameters = hyperparameters
        self.signal_variance = hyperparameters[0]
        self.length_scale = hyperparameters[1]
        self.cutoff = cutoff
        self.cutoff_func = cutoff_func

    def energy_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return energy_energy(*args)

    def force_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return force_energy(*args)

    def stress_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return stress_energy(*args)

    def force_force(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return force_force(*args)

    def stress_force(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return stress_force(*args)

    def stress_stress(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return stress_stress(*args)

    def force_force_gradient(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return force_force_gradient(*args)

    def efs_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return efs_energy(*args)

    def efs_force(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return efs_force(*args)

    def efs_self(self, env1: AtomicEnvironment):
        return efs_self(
            env1.bond_array_2,
            env1.ctype,
            env1.etypes,
            self.signal_variance,
            self.length_scale,
            self.cutoff,
            self.cutoff_func,
        )

    def get_args(self, env1, env2):
        return (
            env1.bond_array_2,
            env1.ctype,
            env1.etypes,
            env2.bond_array_2,
            env2.ctype,
            env2.etypes,
            self.signal_variance,
            self.length_scale,
            self.cutoff,
            self.cutoff_func,
        )


@njit
def energy_energy(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):
    """2-body multi-element kernel between two local energies accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): 2-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 2-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        sig (float): 2-body signal variance hyperparameter.
        ls (float): 2-body length scale hyperparameter.
        r_cut (float): 2-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        float:
            Value of the 2-body local energy kernel.
    """
    kern = 0

    ls1 = 1 / (2 * ls * ls)
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        fi, _ = cutoff_func(r_cut, ri, 0)
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                fj, _ = cutoff_func(r_cut, rj, 0)
                r11 = ri - rj

                kern += fi * fj * sig2 * exp(-r11 * r11 * ls1)

    # Divide by 4 to eliminate double counting (each pair will be counted
    # twice when summing over all environments in the structures).
    return kern / 4


@njit
def force_force(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):
    """2-body multi-element kernel between two force components accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): 2-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 2-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        d1 (int): Force component of the first environment (1=x, 2=y, 3=z).
        d2 (int): Force component of the second environment (1=x, 2=y, 3=z).
        sig (float): 2-body signal variance hyperparameter.
        ls (float): 2-body length scale hyperparameter.
        r_cut (float): 2-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Return:
        float: Value of the 2-body kernel.
    """
    kernel_matrix = np.zeros((3, 3))

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # check if bonds agree
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                r11 = ri - rj
                D = r11 * r11

                # Note: Some redundancy here; can move this higher up.
                for d1 in range(3):
                    ci = bond_array_1[m, d1 + 1]
                    fi, fdi = cutoff_func(r_cut, ri, ci)

                    for d2 in range(3):
                        cj = bond_array_2[n, d2 + 1]
                        fj, fdj = cutoff_func(r_cut, rj, cj)

                        A = ci * cj
                        B = r11 * ci
                        C = r11 * cj

                        kernel_matrix[d1, d2] += force_helper(
                            A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2
                        )

    return kernel_matrix


@njit
def force_energy(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):
    """2-body multi-element kernel between a force component and a local
    energy accelerated with Numba.

    Args:
        bond_array_1 (np.ndarray): 2-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 2-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        d1 (int): Force component of the first environment (1=x, 2=y, 3=z).
        sig (float): 2-body signal variance hyperparameter.
        ls (float): 2-body length scale hyperparameter.
        r_cut (float): 2-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        float:
            Value of the 2-body force/energy kernel.
    """

    kern = np.zeros(3)

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # Check if species agree.
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                fj, _ = cutoff_func(r_cut, rj, 0)
                r11 = ri - rj
                D = r11 * r11

                for d1 in range(3):
                    ci = bond_array_1[m, d1 + 1]
                    fi, fdi = cutoff_func(r_cut, ri, ci)
                    B = r11 * ci
                    kern[d1] += force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2)

    return kern / 2


@njit
def stress_energy(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):
    """2-body multi-element kernel between a partial stress component and a
    local energy accelerated with Numba.

    Args:
        bond_array_1 (np.ndarray): 2-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 2-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        sig (float): 2-body signal variance hyperparameter.
        ls (float): 2-body length scale hyperparameter.
        r_cut (float): 2-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        float:
            Value of the 2-body partial-stress/energy kernel.
    """

    kern = np.zeros(6)

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # Check if the species agree.
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                fj, _ = cutoff_func(r_cut, rj, 0)
                r11 = ri - rj
                D = r11 * r11

                # Compute the force kernel.
                stress_count = 0
                for d1 in range(3):
                    ci = bond_array_1[m, d1 + 1]
                    B = r11 * ci
                    fi, fdi = cutoff_func(r_cut, ri, ci)
                    force_kern = force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2)

                    # Compute the stress kernel from the force kernel.
                    for d2 in range(d1, 3):
                        coordinate = bond_array_1[m, d2 + 1] * ri
                        kern[stress_count] -= force_kern * coordinate
                        stress_count += 1

    return kern / 4


@njit
def stress_force(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):
    """2-body multi-element kernel between two force components accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): 2-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 2-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        sig (float): 2-body signal variance hyperparameter.
        ls (float): 2-body length scale hyperparameter.
        r_cut (float): 2-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Return:
        float: Value of the 2-body kernel.
    """
    kernel_matrix = np.zeros((6, 3))

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # check if bonds agree
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                r11 = ri - rj

                stress_count = 0
                for d1 in range(3):
                    ci = bond_array_1[m, d1 + 1]
                    fi, fdi = cutoff_func(r_cut, ri, ci)
                    for d2 in range(d1, 3):
                        coordinate = bond_array_1[m, d2 + 1] * ri
                        for d3 in range(3):
                            cj = bond_array_2[n, d3 + 1]
                            fj, fdj = cutoff_func(r_cut, rj, cj)

                            A = ci * cj
                            B = r11 * ci
                            C = r11 * cj
                            D = r11 * r11

                            force_kern = force_helper(
                                A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2
                            )
                            kernel_matrix[stress_count, d3] -= force_kern * coordinate

                        stress_count += 1

    return kernel_matrix / 2


@njit
def stress_stress(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):
    """2-body multi-element kernel between two partial stress components
        accelerated with Numba.

    Args:
        bond_array_1 (np.ndarray): 2-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 2-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        sig (float): 2-body signal variance hyperparameter.
        ls (float): 2-body length scale hyperparameter.
        r_cut (float): 2-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Return:
        float: Value of the 2-body kernel.
    """
    kernel_matrix = np.zeros((6, 6))

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # check if bonds agree
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                r11 = ri - rj
                D = r11 * r11

                s1 = 0
                for d1 in range(3):
                    ci = bond_array_1[m, d1 + 1]
                    B = r11 * ci
                    fi, fdi = cutoff_func(r_cut, ri, ci)
                    for d2 in range(d1, 3):
                        coordinate_1 = bond_array_1[m, d2 + 1] * ri
                        s2 = 0
                        for d3 in range(3):
                            cj = bond_array_2[n, d3 + 1]
                            A = ci * cj
                            C = r11 * cj
                            fj, fdj = cutoff_func(r_cut, rj, cj)
                            for d4 in range(d3, 3):
                                coordinate_2 = bond_array_2[n, d4 + 1] * rj
                                force_kern = force_helper(
                                    A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2
                                )
                                kernel_matrix[s1, s2] += (
                                    force_kern * coordinate_1 * coordinate_2
                                )

                                s2 += 1
                        s1 += 1

    return kernel_matrix / 4


@njit
def force_force_gradient(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):
    """2-body multi-element kernel between two force components and its
    gradient with respect to the hyperparameters.

    Args:
        bond_array_1 (np.ndarray): 2-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 2-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        sig (float): 2-body signal variance hyperparameter.
        ls (float): 2-body length scale hyperparameter.
        r_cut (float): 2-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        (float, float):
            Value of the 2-body kernel and its gradient with respect to the
            hyperparameters.
    """

    kernel_matrix = np.zeros((3, 3))
    kernel_grad = np.zeros((2, 3, 3))

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    ls4 = 1 / (ls * ls * ls)
    ls5 = ls * ls
    ls6 = ls2 * ls4

    sig2 = sig * sig
    sig3 = 2 * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # check if bonds agree
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                r11 = ri - rj
                D = r11 * r11

                for d1 in range(3):
                    ci = bond_array_1[m, d1 + 1]
                    B = r11 * ci
                    fi, fdi = cutoff_func(r_cut, ri, ci)

                    for d2 in range(3):
                        cj = bond_array_2[n, d2 + 1]
                        fj, fdj = cutoff_func(r_cut, rj, cj)

                        A = ci * cj
                        C = r11 * cj

                        kern_term, sig_term, ls_term = grad_helper(
                            A,
                            B,
                            C,
                            D,
                            fi,
                            fj,
                            fdi,
                            fdj,
                            ls1,
                            ls2,
                            ls3,
                            ls4,
                            ls5,
                            ls6,
                            sig2,
                            sig3,
                        )

                        kernel_matrix[d1, d2] += kern_term
                        kernel_grad[0, d1, d2] += sig_term
                        kernel_grad[1, d1, d2] += ls_term

    return kernel_matrix, kernel_grad


@njit
def efs_energy(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):

    energy_kernel = 0
    # TODO: add dtype to other zeros
    force_kernels = np.zeros(3)
    stress_kernels = np.zeros(6)

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        fi, _ = cutoff_func(r_cut, ri, 0)
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # Check if the species agree.
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                fj, _ = cutoff_func(r_cut, rj, 0)
                r11 = ri - rj
                D = r11 * r11

                energy_kernel += fi * fj * sig2 * exp(-D * ls1) / 4

                # Compute the force kernel.
                stress_count = 0
                for d1 in range(3):
                    ci = bond_array_1[m, d1 + 1]
                    B = r11 * ci
                    _, fdi = cutoff_func(r_cut, ri, ci)
                    force_kern = force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2)
                    force_kernels[d1] += force_kern / 2

                    # Compute the stress kernel from the force kernel.
                    for d2 in range(d1, 3):
                        coordinate = bond_array_1[m, d2 + 1] * ri
                        stress_kernels[stress_count] -= force_kern * coordinate / 4
                        stress_count += 1

    return energy_kernel, force_kernels, stress_kernels


@njit
def efs_force(
    bond_array_1, c1, etypes1, bond_array_2, c2, etypes2, sig, ls, r_cut, cutoff_func
):

    energy_kernels = np.zeros(3)
    force_kernels = np.zeros((3, 3))
    stress_kernels = np.zeros((6, 3))

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        fi, _ = cutoff_func(r_cut, ri, 0)
        e1 = etypes1[m]

        for n in range(bond_array_2.shape[0]):
            e2 = etypes2[n]

            # check if bonds agree
            if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                rj = bond_array_2[n, 0]
                fj, _ = cutoff_func(r_cut, rj, 0)
                r11 = ri - rj
                D = r11 * r11

                for d1 in range(3):
                    cj = bond_array_2[n, d1 + 1]
                    _, fdj = cutoff_func(r_cut, rj, cj)

                    C = r11 * cj

                    energy_kernels[d1] += (
                        force_energy_helper(-C, D, fj, fi, fdj, ls1, ls2, sig2) / 2
                    )

                    stress_count = 0
                    for d3 in range(3):
                        ci = bond_array_1[m, d3 + 1]
                        _, fdi = cutoff_func(r_cut, ri, ci)
                        A = ci * cj
                        B = r11 * ci

                        force_kern = force_helper(
                            A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2
                        )
                        force_kernels[d3, d1] += force_kern

                        for d2 in range(d3, 3):
                            coordinate = bond_array_1[m, d2 + 1] * ri
                            stress_kernels[stress_count, d1] -= (
                                force_kern * coordinate / 2
                            )

                            stress_count += 1

    return energy_kernels, force_kernels, stress_kernels


@njit
def efs_self(bond_array_1, c1, etypes1, sig, ls, r_cut, cutoff_func):

    energy_kernel = 0
    force_kernels = np.zeros(3)
    stress_kernels = np.zeros(6)

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig * sig

    for m in range(bond_array_1.shape[0]):
        ri = bond_array_1[m, 0]
        fi, _ = cutoff_func(r_cut, ri, 0)
        e1 = etypes1[m]

        for n in range(bond_array_1.shape[0]):
            e2 = etypes1[n]

            # check if bonds agree
            if (e1 == e2) or (c1 == e2 and c1 == e1):
                rj = bond_array_1[n, 0]
                fj, _ = cutoff_func(r_cut, rj, 0)
                r11 = ri - rj
                D = r11 * r11

                energy_kernel += fi * fj * sig2 * exp(-D * ls1) / 4

                stress_count = 0
                for d1 in range(3):
                    cj = bond_array_1[n, d1 + 1]
                    _, fdj = cutoff_func(r_cut, rj, cj)
                    C = r11 * cj

                    ci = bond_array_1[m, d1 + 1]
                    _, fdi = cutoff_func(r_cut, ri, ci)
                    A = ci * cj
                    B = r11 * ci

                    force_kern = force_helper(
                        A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2
                    )
                    force_kernels[d1] += force_kern

                    for d2 in range(d1, 3):
                        coord1 = bond_array_1[m, d2 + 1] * ri
                        coord2 = bond_array_1[n, d2 + 1] * rj
                        stress_kernels[stress_count] += force_kern * coord1 * coord2 / 4

                        stress_count += 1

    return energy_kernel, force_kernels, stress_kernels



"""
This module is used to call CP2K simulation and parse its output
The user need to supply a complete input script with ENERGY_FORCE or ENERGY runtype, and CELL, COORD blocks. Example scripts can be found in tests/test_files/cp2k_input...

The module will copy the input template to a new file with "_run" suffix,
edit the atomic coordination in the COORD blocks and run the similation with
the parallel set up given.

We note that, if the CP2K executable is only for serial run, using it along with MPI setting can lead to repeating output in the output file, wrong number of forces and error in the other modules.
"""

import os
from subprocess import call
import time
import numpy as np
from flare import output
from flare import struc
from typing import List

name = "CP2K"


def run_dft_par(
    dft_input,
    structure,
    dft_loc,
    ncpus=1,
    dft_out="dft.out",
    npool=None,
    mpi="mpi",
    **dft_kwargs,
):
    """run DFT calculation with given input template
    and atomic configurations. if ncpus == 1, it executes serial run.

    :param dft_input: input template file name
    :param structure: atomic configuration
    :param dft_loc:   relative/absolute executable of the DFT code
    :param ncpus:   # of CPU for mpi
    :param dft_out:   output file name
    :param npool:     not used
    :param mpi:       not used
    :param **dft_wargs: not used
    :return: forces
    """

    newfilename = edit_dft_input_positions(dft_input, structure)
    dft_command = f"{dft_loc} -i {newfilename}"
    if ncpus > 1:
        if mpi == "mpi":
            dft_command = f"mpirun -np {ncpus} {dft_command}"
        else:
            dft_command = f"srun -n {ncpus} {dft_command}"

    # output.write_to_output(dft_command+'\n')
    with open(dft_out, "w+") as fout:
        call(dft_command.split(), stdout=fout)

    os.remove(newfilename)

    return parse_dft_forces(dft_out)


def run_dft_en_par(
    dft_input: str,
    structure,
    dft_loc: str,
    ncpus: int,
    dft_out: str = "dft.out",
    npool: int = None,
    mpi: str = "mpi",
    **dft_kwargs,
):
    """run DFT calculation with given input template
    and atomic configurations. This function is not used atm.

    :param dft_input: input template file name
    :param structure: atomic configuration
    :param dft_loc:   relative/absolute executable of the DFT code
    :param ncpus:   # of CPU for mpi
    :param dft_out:   output file name
    :param npool:     not used
    :param mpi:       not used
    :param **dft_wargs: not used
    :return: forces, energy
    """

    newfilename = edit_dft_input_positions(dft_input, structure)
    dft_command = f"{dft_loc} -i {newfilename} > {dft_out}"
    if ncpus > 1:
        dft_command = f"mpirun -np {ncpus} {dft_command}"
    # output.write_to_output(dft_command+'\n')
    call(dft_command, shell=True)
    os.remove(newfilename)

    forces, energy = parse_dft_forces_and_energy(dft_out)

    return forces, energy


def parse_dft_input(dft_input: str):
    """Parse CP2K input file prepared by the user
    the parser is very limited. The user have to define things
    in a good format.
    It requires the "CELL", "COORD" blocks

    :param dft_input: file name
    :return: positions, species, cell, masses
    """
    positions = []
    species = []
    cell = []

    with open(dft_input) as f:
        lines = f.readlines()

    # Find the cell and positions in the output file
    cell_index = None
    positions_index = None
    nat = None
    # species_index = None

    for i, line in enumerate(lines):
        if "&CELL" in line:
            cell_index = int(i + 1)
        elif "COORD" in line and "END" not in line:
            positions_index = int(i + 1)
        elif "&END" in line and (positions_index is not None) and (nat is None):
            nat = i - positions_index
    #     if 'ATOMIC_SPECIES' in line:
    #         species_index = int(i + 1)

    assert cell_index is not None, "Failed to find cell in input"
    assert positions_index is not None, "Failed to find positions in input"
    assert nat is not None, "Failed to find number of atoms in input"

    # Load cell
    # TO DO: allow to mess up the order of A, B, and C
    for i in range(cell_index, cell_index + 3):
        cell_line = list(map(float, lines[i].split()[1:]))
        cell.append(cell_line)  # np.fromstring(cell_line[1:], sep=' '))
    cell = np.array(cell)

    # Check cell IO
    assert len(cell) != 0, "Cell failed to load"
    assert np.shape(cell) == (3, 3), "Cell failed to load correctly"

    # Load positions
    for i in range(positions_index, positions_index + nat):
        pos_line = lines[i].split()
        species.append(pos_line[0])
        # positions.append(np.fromstring(pos_string, sep=' '))
        positions.append(list(map(float, pos_line[1:])))

    # Check position IO
    assert positions != [], "Positions failed to load"
    positions = np.array(positions)

    # see conversions.nb for conversion from amu to md units
    ele_mass = {
        "H": 1.007900,
        "He": 4.002600,
        "Li": 6.941000,
        "Be": 9.012200,
        "B": 10.811000,
        "C": 12.010700,
        "N": 14.006700,
        "O": 15.999400,
        "F": 18.998400,
        "Ne": 20.179700,
        "Na": 22.989700,
        "Mg": 24.305000,
        "Al": 26.981500,
        "Si": 28.085500,
        "P": 30.973800,
        "S": 32.065000,
        "Cl": 35.453000,
        "K": 39.098300,
        "Ar": 39.948000,
        "Ca": 40.078000,
        "Sc": 44.955900,
        "Ti": 47.867000,
        "V": 50.941500,
        "Cr": 51.996100,
        "Mn": 54.938000,
        "Fe": 55.845000,
        "Ni": 58.693400,
        "Co": 58.933200,
        "Cu": 63.546000,
        "Zn": 65.390000,
        "Ga": 69.723000,
        "Ge": 72.640000,
        "As": 74.921600,
        "Se": 78.960000,
        "Br": 79.904000,
        "Kr": 83.800000,
        "Rb": 85.467800,
        "Sr": 87.620000,
        "Y": 88.905900,
        "Zr": 91.224000,
        "Nb": 92.906400,
        "Mo": 95.940000,
        "Tc": 98.000000,
        "Ru": 101.070000,
        "Rh": 102.905500,
        "Pd": 106.420000,
        "Ag": 107.868200,
        "Cd": 112.411000,
        "In": 114.818000,
        "Sn": 118.710000,
        "Sb": 121.760000,
        "I": 126.904500,
        "Te": 127.600000,
        "Xe": 131.293000,
        "Cs": 132.905500,
        "Ba": 137.327000,
        "La": 138.905500,
        "Ce": 140.116000,
        "Pr": 140.907700,
        "Nd": 144.240000,
        "Pm": 145.000000,
        "Sm": 150.360000,
        "Eu": 151.964000,
        "Gd": 157.250000,
        "Tb": 158.925300,
        "Dy": 162.500000,
        "Ho": 164.930300,
        "Er": 167.259000,
        "Tm": 168.934200,
        "Yb": 173.040000,
        "Lu": 174.967000,
        "Hf": 178.490000,
        "Ta": 180.947900,
        "W": 183.840000,
        "Re": 186.207000,
        "Os": 190.230000,
        "Ir": 192.217000,
        "Pt": 195.078000,
        "Au": 196.966500,
        "Hg": 200.590000,
        "Tl": 204.383300,
        "Pb": 207.200000,
        "Bi": 208.980400,
        "Po": 209.000000,
        "At": 210.000000,
        "Rn": 222.000000,
        "Fr": 223.000000,
        "Ra": 226.000000,
        "Ac": 227.000000,
        "Pa": 231.035900,
        "Th": 232.038100,
        "Np": 237.000000,
        "U": 238.028900,
        "Am": 243.000000,
        "Pu": 244.000000,
        "Cm": 247.000000,
        "Bk": 247.000000,
        "Cf": 251.000000,
        "Es": 252.000000,
        "Fm": 257.000000,
        "Md": 258.000000,
        "No": 259.000000,
        "Rf": 261.000000,
        "Lr": 262.000000,
        "Db": 262.000000,
        "Bh": 264.000000,
        "Sg": 266.000000,
        "Mt": 268.000000,
        "Rg": 272.000000,
        "Hs": 277.000000,
    }

    # TO DO: allow customize mass
    massconvert = 0.000103642695727
    masses = {}
    for ele in ele_mass.keys():
        # Expects lines of format like: H 1.0 H_pseudo_name.ext
        masses[ele] = ele_mass[ele] * massconvert

    return positions, species, cell, masses


def dft_input_to_structure(dft_input: str):
    """
    Parses a qe input and returns the atoms in the file as a Structure object
    :param dft_input: input file to parse
    :return: atomic structure
    """
    positions, species, cell, masses = parse_dft_input(dft_input)
    _, coded_species = struc.get_unique_species(species)
    return struc.Structure(
        positions=positions,
        species=coded_species,
        cell=cell,
        mass_dict=masses,
        species_labels=species,
    )


def edit_dft_input_positions(dft_input: str, structure):
    """Write the current configuration of the OTF structure to the
    qe input file

    :param dft_input: intput file name
    :param structure: structure to print
    :type structure: class Structure
    :return newfilename: the name of the edited intput file.
                         with "_run" suffix
    """

    with open(dft_input, "r") as f:
        lines = f.readlines()

    file_pos_index = None
    cell_index = None
    nat = None

    for i, line in enumerate(lines):
        if "&CELL" in line:
            cell_index = int(i + 1)
        if "&COORD" in line:
            file_pos_index = int(i + 1)
        if "&END" in line and (file_pos_index is not None):
            nat = i - file_pos_index
    #     if 'ATOMIC_SPECIES' in line:
    #         species_index = int(i + 1)

    assert file_pos_index is not None, "Failed to find positions in input"
    assert cell_index is not None, "Failed to find cell in input"
    assert nat is not None, "Failed to find nat in input"

    for pos_index, line_index in enumerate(
        range(file_pos_index, file_pos_index + structure.nat)
    ):
        pos = structure.positions[pos_index]
        specs = structure.species_labels[pos_index]
        pos_string = f"{specs} {pos[0]} {pos[1]} {pos[2]}\n"
        if line_index < len(lines):
            lines[line_index] = pos_string
        else:
            lines.append(pos_string)

    # # TODO current assumption: if there is a new structure, then the new
    # # structure has fewer atoms than the  previous one. If we are always
    # # 'editing' a version of the larger structure than this will be okay with
    # # the punchout method.
    # for line_index in range(file_pos_index + structure.nat,
    #                         file_pos_index + nat):
    #     lines[line_index] = ''

    lines[cell_index] = "A " + " ".join([str(x) for x in structure.vec1]) + "\n"
    lines[cell_index + 1] = "B " + " ".join([str(x) for x in structure.vec2]) + "\n"
    lines[cell_index + 2] = "C " + " ".join([str(x) for x in structure.vec3]) + "\n"

    newfilename = dft_input + "_run"

    with open(newfilename, "w") as f:
        for line in lines:
            f.write(line)

    return newfilename


def parse_dft_forces_and_energy(outfile: str):
    """Get forces from a pwscf file in eV/A
    the input run type to be ENERGY_FORCE

    :param outfile: str, Path to dft.output file
    :return: list[nparray] , List of forces acting on atoms
    :return: float, total potential energy
    """
    forces = []
    total_energy = np.nan

    startforce = -1
    with open(outfile, "r") as outf:
        for line in outf:
            if line.find("FORCE_EVAL") != -1:
                total_energy = float(line.split()[8])

            if startforce >= 2:
                if line.find("SUM") != -1:
                    startforce = -1
                else:
                    line = line.split()[3:]
                    forces.append(list(map(float, line)))
                    startforce += 1
            elif startforce >= 0:
                startforce += 1
            elif line.find("FORCES") != -1 and line.find("in") != -1:
                startforce = 0

    assert total_energy != np.nan, (
        "dft parser failed to read the file {}. Run failed." + outfile
    )

    # Convert from ry/au to ev/angstrom
    conversion_factor = 25.71104309541616 * 2.0

    forces = np.array(forces) * conversion_factor
    total_energy *= 27.2114

    return forces, total_energy


def parse_dft_forces(outfile: str):
    """Get forces from a pwscf file in eV/A

    :param outfile: str, Path to dft.output file
    :return: list[nparray] , List of forces acting on atoms
    """

    f, e = parse_dft_forces_and_energy(outfile)
    return f



import numpy as np
from numba import njit
from math import exp, floor
from typing import Callable

from flare.kernels.cutoffs import quadratic_cutoff

from time import time


def grid_kernel(struc, *args):

    if not isinstance(struc, list):
        struc = [struc]

    kern = 0
    for env in struc:
        kern += grid_kernel_env(env, *args)

    return kern


def grid_kernel_env(
    env1,
    bodies,
    kern_type,
    get_bonds_func,
    bonds_cutoff_func,
    c2,
    etypes2,
    hyps,
    r_cut,
    grids,
    fj,
    fdj,
    cutoff_func: Callable = quadratic_cutoff,
):

    # pre-compute constants that appear in the inner loop
    sig = hyps[0]
    ls = hyps[1]
    derivative = derv_dict[kern_type]
    grid_dim = grids.shape[1]

    # collect all the triplets in this training env
    bond_coord_list = get_bonds_func(env1, c2, etypes2)

    if len(bond_coord_list) == 0:  # no triplets
        if derivative:
            return np.zeros((3, grids.shape[0]), dtype=np.float64)
        else:
            return np.zeros(grids.shape[0], dtype=np.float64)

    bond_coord_list = np.array(bond_coord_list)
    bond_list = bond_coord_list[:, :grid_dim]
    coord_list = bond_coord_list[:, grid_dim:]
    del bond_coord_list

    # calculate distance difference & exponential part
    ls1 = 1 / (2 * ls * ls)
    D = 0
    rij_list = []
    for r in range(grid_dim):
        rj, ri = np.meshgrid(grids[:, r], bond_list[:, r])
        rij = ri - rj
        D += rij * rij  # (n_triplets, n_grids)
        rij_list.append(rij)

    kern_exp = (sig * sig) * np.exp(-D * ls1)
    del D

    # calculate cutoff of the triplets
    fi, fdi = bonds_cutoff_func(
        bond_list, r_cut, coord_list, derivative, cutoff_func
    )  # (n_triplets, 1)
    del bond_list

    # calculate the derivative part
    kern_func = kern_dict[kern_type]
    kern = kern_func(
        bodies, grid_dim, kern_exp, fi, fj, fdi, fdj, rij_list, coord_list, ls
    )

    return kern


def en_en(bodies, grid_dim, kern_exp, fi, fj, *args):
    """energy map + energy block"""
    fifj = fi @ fj.T  # (n_triplets, n_grids)
    kern = np.sum(kern_exp * fifj, axis=0) / bodies ** 2  # (n_grids,)
    return kern


def en_force(bodies, grid_dim, kern_exp, fi, fj, fdi, fdj, rij_list, coord_list, ls):
    """energy map + force block"""
    fifj = fi @ fj.T  # (n_triplets, n_grids)
    ls2 = 1 / (ls * ls)
    n_trplt, n_grids = kern_exp.shape
    kern = np.zeros((3, n_grids), dtype=np.float64)
    for d in range(3):
        B = 0
        fdij = fdi[:, [d]] @ fj.T
        for r in range(grid_dim):
            rij = rij_list[r]
            # column-wise multiplication
            # coord_list[:, [r]].shape = (n_triplets, 1)
            B += rij * coord_list[:, [3 * r + d]]  # (n_triplets, n_grids)

        kern[d, :] = (
            -np.sum(kern_exp * (B * ls2 * fifj + fdij), axis=0) / bodies
        )  # (n_grids,)
    return kern


def self_kernel(
    bodies,
    get_permutations,
    c2,
    etypes2,
    hyps,
    r_cut,
    grids,
    fj,
    fdj,
    cutoff_func: Callable = quadratic_cutoff,
):

    kern = 0

    # pre-compute constants
    sig = hyps[0]
    ls = hyps[1]
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    perm_list = get_permutations(c2, etypes2, c2, etypes2)

    for perm in perm_list:
        perm_grids = np.take(grids, perm, axis=1)
        rij = grids - perm_grids
        D = np.sum(rij * rij, axis=1)  # (n_grids, ) adding up three bonds
        kern_exp = np.exp(-D * ls1) * sig2
        fjfj = fj ** 2
        kern += kern_exp * np.sum(fjfj, axis=1) / bodies ** 2  # (n_grids,)

    return kern


kern_dict = {
    "energy_energy": en_en,
    "energy_force": en_force,
}

derv_dict = {
    "energy_energy": False,
    "energy_force": True,
}
